R"""


cd ~/Desktop/projects/extract_merge1
export PYTHONPATH=$PYTHONPATH:~/Desktop/projects/extract_merge1


CUDA_VISIBLE_DEVICES=0 python -i local_scripts/ll/hans_merging_04.py

"""

import dataclasses
from importlib import reload
import itertools
import os
import pickle
import time

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import tensorflow as tf
from tqdm import tqdm
from transformers import AutoTokenizer, TFAutoModelForSequenceClassification

from em import datasets as em_datasets
from em.fishers import diagonal
from em.fishers import per_example
from em.merging import merging
from em.models import transformer_model_vars as tmv
from em.projects.anli import anli_misc1 as am
from em.projects.ll import hans_util
from em.projects.ll import hans_analysis as ha
from em.projects.wino import nmf_components_fisher as ncf
from em.tools.clustering import vat
from em.projects.ll import hans_labeling
from em.projects.ll import hans_labeling_analysis as hla
from em.tools.nmf import nmf_common
from em.util import flat_pack
from em.util import hf_util

###############################################################################

EXPS_DIR = '/fruitbasket/users/m/project_data/extract_merge1/ll1'
MODELS_DIR = os.path.join(EXPS_DIR, 'models')
FISHERS_DIR = os.path.join(EXPS_DIR, 'fishers')
PER_EXAMPLES_FISHERS_DIR = os.path.join(EXPS_DIR, 'per_example_fishers')

###############################################################################

TOKENIZER = 'bert-base-uncased'

###############################################################################

HANS_LONE_PEF_FILENAME = "feather_berts_{model_number}.hans_lone.all_vars.5k.262144.h5"
HANS_LONE_NMF_FILENAME = "spH.nmf_decomp.c{n_components}_2kIters_{n_fisher_values}pe.{pef}"

FISHER_FILENAME = "feather_berts_{model_number}.hans_lone.validation.h5"

##########################################################################

tokenizer = AutoTokenizer.from_pretrained(TOKENIZER)

examples = hans_util.get_first_hans_examples(
    'validation',
    5000,
    lambda ds: ds.filter(lambda x: x['heuristic'] == 'lexical_overlap' and x['label'] == 1)
)

indicators = hans_labeling.compute_full_indicator(examples, 'ne')

##########################################################################


def make_container(model_number: int, n_components: int = 256, n_fisher_values: int = 65536):
    pef_file = HANS_LONE_PEF_FILENAME.format(model_number=model_number)
    nmf_file = HANS_LONE_NMF_FILENAME.format(n_components=n_components, pef=pef_file, n_fisher_values=n_fisher_values)
    #
    pef = per_example.PerExampleFlatFishers.load(
        os.path.join(PER_EXAMPLES_FISHERS_DIR, pef_file),
        n_examples=None,
        # This leads to the Fishers not being loaded, which ends up being much faster.
        start_fisher_index=0,
        end_fisher_index=0,
    )
    nmf = nmf_common.SparseNmfDecomposition.load(os.path.join(PER_EXAMPLES_FISHERS_DIR, nmf_file))
    nmf.normalize_components_to_unit_norm()
    #
    container = am.PefNmfAnalysisContainer(
        pef=pef,
        nmfs=[nmf],
        tokenizer=tokenizer,
        shift_labels=False,
    )
    hans_util.fix_up_hans_container(container)
    return container


def get_model(model_number: int):
    model = TFAutoModelForSequenceClassification.from_pretrained(
        f'connectivity/feather_berts_{model_number}', from_pt=True)
    model.compile(
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        metrics=[tf.keras.metrics.SparseCategoricalAccuracy()],
    )
    return model


def get_fisher(model_number: int):
    filename = FISHER_FILENAME.format(model_number=model_number)
    return diagonal.DiagonalFisher.load(os.path.join(FISHERS_DIR, filename))


##########################################################################

model_number_a = 0
model_number_b = 15

container_a = make_container(model_number_a, 1024, 65536 * 2)
model_a = get_model(model_number_a)
dense_fisher_a = get_fisher(model_number_a).fishers
print("Loaded all of A.")

container_b = make_container(model_number_b, 1024, 65536 * 2)
model_b = get_model(model_number_b)
dense_fisher_b = get_fisher(model_number_b).fishers
print("Loaded all of B.")

##########################################################################

selection_parameters = ncf.SelectionParameters(
    coeff_factor=0.3,
    frac_threshold=0.925,
    p_value_threshold=0.15,
    # 
    # # coeff_factor=1 / 3,
    # coeff_factor=0.4,
    # frac_threshold=0.85,
    # # frac_threshold=2 / 3,
    # # frac_threshold=0.5,
    # # p_value_threshold=0.05,
    # p_value_threshold=0.15,
)

tuning_info_a = hla.compute_hans_tuning_info(container_a, indicators, selection_parameters)
tuning_info_b = hla.compute_hans_tuning_info(container_b, indicators, selection_parameters)

##########################################################################

ccsp = ncf.SelectionParameters(
    # coeff_factor=0.4,
    # frac_threshold=0.85,
    # p_value_threshold=0.15,
    # 
    coeff_factor=0.3,
    frac_threshold=0.925,
    p_value_threshold=0.15,
)

correct_comp_infos_a = ncf.get_components_appearing_tuned(
    container_a,
    indicator=container_a.get_correct_prediction_indicator(),
    selection_parameters=ccsp,
)
correct_comp_infos_b = ncf.get_components_appearing_tuned(
    container_b,
    indicator=container_b.get_correct_prediction_indicator(),
    selection_parameters=ccsp,
)
##########################################################################


def make_sparse_fisher_vector_for_components(tuning_info, component_indices) -> tf.sparse.SparseTensor:
    component_indices = np.array(list(sorted(component_indices)), dtype=np.int32)
    #
    nmf, = tuning_info.container.nmfs
    spH = nmf.get_full_sparse_H()
    spH = [spH[i] for i in component_indices]
    #
    # W.shape = [n_examples, len(component_indices)]
    W = hla._get_denormalized_coefficients(tuning_info.container, 0, component_indices)
    coeffs = W.sum(axis=0)
    #
    terms = [tf.cast(c, tf.float32) * h for c, h in zip(coeffs, spH)]
    ret = terms[0]
    for term in terms[1:]:
        ret = tf.sparse.add(ret, term)
    return ret


def make_fisher_for_components(tuning_info, component_indices, variables):
    spF = make_sparse_fisher_vector_for_components(tuning_info, component_indices)
    packer = flat_pack.FlatPacker([v.shape for v in variables])
    return packer.decode_tf(tf.sparse.to_dense(spF))


def get_tuned_component_indices(tuning_info, key):
    tcis_by_subset, = hla._get_item_tupled_key(tuning_info.tcis_by_nmf_dict, key)
    return [t.component_index for t in tcis_by_subset]


def get_not_tuned_component_indices(tuning_info, key):
    tuned_inds = get_tuned_component_indices(tuning_info, key)
    n_components = tuning_info.container.nmfs[0].W.shape[-1]
    return list(sorted(set(range(n_components)) - set(tuned_inds)))


##########################################################################

example_type = ('subcase_indicators', 'ln_preposition')
# example_type = ('subcase_indicators', 'ln_conjunction')
# example_type = ('subcase_indicators', 'ln_subject/object_swap')


fisher_a = make_fisher_for_components(
    tuning_info_a,
    # get_not_tuned_component_indices(tuning_info_a, example_type),
    list(range(tuning_info_a.container.nmfs[0].W.shape[-1])),
    # get_tuned_component_indices(tuning_info_a, example_type),
    hf_util.get_all_variables(model_a),
)
fisher_b = make_fisher_for_components(
    tuning_info_b,
    get_tuned_component_indices(tuning_info_b, example_type),
    hf_util.get_all_variables(model_b),
)

# fisher_a = dense_fisher_a
# fisher_b = dense_fisher_b


##########################################################################

eval_batch_size = 128

val_ds = em_datasets.load('hans/lexical_overlap_ne', split='validation', sequence_length=64, tokenizer=tokenizer)


def evaluate_model(output_model):
    labels = []
    preds = []
    for x, y in val_ds.batch(eval_batch_size):
        logits = output_model(x, training=False).logits.numpy()
        logits = hans_util.fix_up_hans_logits(logits)
        batch_preds = np.argmax(logits, axis=-1)
        labels.append(y.numpy())
        preds.append(batch_preds)
    labels = np.concatenate(labels, axis=0)
    preds = np.concatenate(preds, axis=0)
    #
    for key, value in indicators.subcase_indicators.items():
        corrects = preds == labels
        acc = corrects[value].astype(np.float64).mean()
        print(f'{key}: {acc}')


##########################################################################

variables_a = hf_util.get_all_variables(model_a)
variables_b = hf_util.get_all_variables(model_b)

output_model = get_model(model_number_a)

output_variables = hf_util.get_all_variables(output_model)
variables_to_merge = [variables_a, variables_b]
fishers_to_merge = [fisher_a, fisher_b]

norm_constants = [merging._l2_norm_of_fisher(f) for f in fishers_to_merge]

# gen = merging.generate_merged_for_coeffs_set(
#     mergeable_models=[model_b, model_a],
#     coefficients_set=merging.create_pairwise_grid_coeffs(21),
#     fishers=[fisher_b, fisher_a],
#     fisher_floor=1e-8,
#     favor_target_model=True,
#     normalize_fishers=True,
#     output_model=output_model,
# )
# for coefficients, _ in gen:
#     print(coefficients)
#     evaluate_model(output_model)
#     print('')

for coefficients in merging.create_pairwise_grid_coeffs(21):
# for coefficients in merging.create_pairwise_grid_coeffs(201)[:21]:
# for coefficients in merging.create_pairwise_grid_coeffs(41)[3:21]:
    merging._merge_with_coeffs(
        output_variables,
        variables_to_merge,
        coefficients=coefficients,
        fishers=fishers_to_merge,
        fisher_floor=1e-8,
        # favor_target_model=True,
        favor_target_model=False,
        normalization_constants=norm_constants,
    )
    print(coefficients)
    evaluate_model(output_model)
    print('')

